@@ -12,6 +12,7 @@
 #include <linux/ipc.h>
 #include <linux/net.h>
 #include <linux/sched.h>
+#include <linux/sockios.h>
 #include <string.h>
 #include <sys/ioctl.h>
 #include <sys/mman.h>
@@ -22,11 +23,14 @@
 #include <sys/utsname.h>
 #include <time.h>
 #include <unistd.h>
+// This has to go after <sys/socket.h> for annoying reasons
+#include <linux/wireless.h>
 
 #include <algorithm>
 #include <utility>
 #include <vector>
 
+#include "PlatformMacros.h"
 #include "Sandbox.h"  // for ContentProcessSandboxParams
 #include "SandboxBrokerClient.h"
 #include "SandboxFilterUtil.h"
@@ -43,6 +47,11 @@
 #include "sandbox/linux/system_headers/linux_seccomp.h"
 #include "sandbox/linux/system_headers/linux_syscalls.h"
 
+#if defined(GP_PLAT_amd64_linux) && defined(GP_ARCH_amd64) && \
+    defined(MOZ_USING_WASM_SANDBOXING)
+#  include <asm/prctl.h>  // For ARCH_SET_GS
+#endif
+
 using namespace sandbox::bpf_dsl;
 #define CASES SANDBOX_BPF_DSL_CASES
 
@@ -100,6 +109,22 @@ static_assert(F_ADD_SEALS == (F_LINUX_SPECIFIC_BASE + 9));
 static_assert(F_GET_SEALS == (F_LINUX_SPECIFIC_BASE + 10));
 #endif
 
+// Added in 6.13
+#ifndef MADV_GUARD_INSTALL
+#  define MADV_GUARD_INSTALL 102
+#  define MADV_GUARD_REMOVE 103
+#else
+static_assert(MADV_GUARD_INSTALL == 102);
+static_assert(MADV_GUARD_REMOVE == 103);
+#endif
+
+// Added in 6.10
+#ifndef F_DUPFD_QUERY
+#  define F_DUPFD_QUERY (F_LINUX_SPECIFIC_BASE + 3)
+#else
+static_assert(F_DUPFD_QUERY == (F_LINUX_SPECIFIC_BASE + 3));
+#endif
+
 // To avoid visual confusion between "ifdef ANDROID" and "ifndef ANDROID":
 #ifndef ANDROID
 #  define DESKTOP
@@ -524,23 +549,65 @@ class SandboxPolicyCommon : public SandboxPolicyBase {
     return ConvertError(socketpair(AF_UNIX, SOCK_SEQPACKET, 0, fds));
   }
 
-  static intptr_t SocketpairUnpackTrap(ArgsRef aArgs, void* aux) {
-#ifdef __NR_socketpair
-    auto argsPtr = reinterpret_cast<unsigned long*>(aArgs.args[1]);
-    return DoSyscall(__NR_socketpair, argsPtr[0], argsPtr[1], argsPtr[2],
-                     argsPtr[3]);
-#else
-    MOZ_CRASH("unreachable?");
-    return -ENOSYS;
-#endif
-  }
+  static intptr_t SocketcallUnpackTrap(ArgsRef aArgs, void* aux) {
+#ifdef __NR_socketcall
+    auto argsPtr = reinterpret_cast<const unsigned long*>(aArgs.args[1]);
+    int sysno = -1;
 
-  static intptr_t GetSockOptUnpackTrap(ArgsRef aArgs, void* aux) {
-#ifdef __NR_getsockopt
-    auto argsPtr = reinterpret_cast<unsigned long*>(aArgs.args[1]);
-    return DoSyscall(__NR_getsockopt, argsPtr[0], argsPtr[1], argsPtr[2],
-                     argsPtr[3], argsPtr[4]);
-#else
+    // When Linux added separate syscalls for socket operations on the
+    // old socketcall platforms, they had long since stopped adding
+    // send and recv syscalls, because they can be trivially mapped
+    // onto sendto and recvfrom (see also open vs. openat).
+    //
+    // But, socketcall itself *does* have separate calls for those.
+    // So, we need to remap them; since send(to) and recv(from)
+    // have basically the same types except for const, the code is
+    // factored out here.
+    unsigned long altArgs[6];
+    auto legacySendRecvWorkaround = [&] {
+      MOZ_ASSERT(argsPtr != altArgs);
+      memcpy(altArgs, argsPtr, sizeof(unsigned long[4]));
+      altArgs[4] = altArgs[5] = 0;
+      argsPtr = altArgs;
+    };
+
+    switch (aArgs.args[0]) {
+      // See also the other socketcall table in SandboxFilterUtil.cpp
+#  define DISPATCH_SOCKETCALL(this_sysno, this_call) \
+    case this_call:                                  \
+      sysno = this_sysno;                            \
+      break
+
+      DISPATCH_SOCKETCALL(__NR_socketpair, SYS_SOCKETPAIR);
+      DISPATCH_SOCKETCALL(__NR_getsockopt, SYS_GETSOCKOPT);
+      DISPATCH_SOCKETCALL(__NR_sendmsg, SYS_SENDMSG);
+      DISPATCH_SOCKETCALL(__NR_recvmsg, SYS_RECVMSG);
+      DISPATCH_SOCKETCALL(__NR_sendto, SYS_SENDTO);
+      DISPATCH_SOCKETCALL(__NR_recvfrom, SYS_RECVFROM);
+      DISPATCH_SOCKETCALL(__NR_sendmmsg, SYS_SENDMMSG);
+      DISPATCH_SOCKETCALL(__NR_recvmmsg, SYS_RECVMMSG);
+      // __NR_recvmmsg_time64 is not available as a socketcall; a
+      // Y2K38-ready userland would call it directly.
+#  undef DISPATCH_SOCKETCALL
+
+      case SYS_SEND:
+        sysno = __NR_sendto;
+        legacySendRecvWorkaround();
+        break;
+      case SYS_RECV:
+        sysno = __NR_recvfrom;
+        legacySendRecvWorkaround();
+        break;
+    }
+
+    // This assert will fail if someone tries to map a socketcall to
+    // this trap without adding it to the switch statement above.
+    MOZ_RELEASE_ASSERT(sysno >= 0);
+
+    return DoSyscall(sysno, argsPtr[0], argsPtr[1], argsPtr[2], argsPtr[3],
+                     argsPtr[4], argsPtr[5]);
+
+#else  // no socketcall
     MOZ_CRASH("unreachable?");
     return -ENOSYS;
 #endif
@@ -737,23 +804,69 @@ class SandboxPolicyCommon : public SandboxPolicyBase {
         .CASES((PR_CAPBSET_READ),  // libcap.so.2 loaded by libpulse.so.0
                                    // queries for capabilities
                Error(EINVAL))
+#if defined(MOZ_PROFILE_GENERATE)
+        .CASES((PR_GET_PDEATHSIG), Allow())
+#endif  // defined(MOZ_PROFILE_GENERATE)
         .Default(InvalidSyscall());
   }
 
+  virtual BoolExpr MsgFlagsAllowed(const Arg<int>& aFlags) const {
+    // MSG_DONTWAIT: used by IPC
+    // MSG_NOSIGNAL: used by the sandbox (broker, reporter)
+    // MSG_CMSG_CLOEXEC: should be used by anything that's passed fds
+    static constexpr int kNeeded =
+        MSG_DONTWAIT | MSG_NOSIGNAL | MSG_CMSG_CLOEXEC;
+
+    // These don't appear to be used in our code at the moment, but
+    // they seem low-risk enough to allow to avoid the possibility of
+    // breakage.  (Necko might use MSG_PEEK, but the socket process
+    // overrides this method.)
+    static constexpr int kHarmless = MSG_PEEK | MSG_WAITALL | MSG_TRUNC;
+
+    static constexpr int kAllowed = kNeeded | kHarmless;
+    return (aFlags & ~kAllowed) == 0;
+  }
+
+  static ResultExpr UnpackSocketcallOrAllow() {
+    // See bug 1066750.
+    if (HasSeparateSocketCalls()) {
+      // If this is a socketcall(2) platform, but the kernel also
+      // supports separate syscalls (>= 4.3.0), we can unpack the
+      // arguments and filter them.
+      return Trap(SocketcallUnpackTrap, nullptr);
+    }
+    // Otherwise, we can't filter the args if the platform passes
+    // them by pointer.
+    return Allow();
+  }
+
   Maybe<ResultExpr> EvaluateSocketCall(int aCall,
                                        bool aHasArgs) const override {
     switch (aCall) {
       case SYS_RECVMSG:
       case SYS_SENDMSG:
-        // These next four aren't needed for IPC or other core
-        // functionality at the time of this writing, but they're
-        // subsets of recvmsg/sendmsg so there's nothing gained by not
-        // allowing them here (and simplifying subclasses).
+        if (aHasArgs) {
+          Arg<int> flags(2);
+          return Some(
+              If(MsgFlagsAllowed(flags), Allow()).Else(InvalidSyscall()));
+        }
+        return Some(UnpackSocketcallOrAllow());
+
+        // These next four weren't needed for IPC or other core
+        // functionality when they were added, but they're subsets of
+        // recvmsg/sendmsg so there's nothing gained by not allowing
+        // them here (and simplifying subclasses).  Also, there may be
+        // unknown dependencies on them now.
       case SYS_RECVFROM:
       case SYS_SENDTO:
       case SYS_RECV:
       case SYS_SEND:
-        return Some(Allow());
+        if (aHasArgs) {
+          Arg<int> flags(3);
+          return Some(
+              If(MsgFlagsAllowed(flags), Allow()).Else(InvalidSyscall()));
+        }
+        return Some(UnpackSocketcallOrAllow());
 
       case SYS_SOCKETPAIR: {
         // We try to allow "safe" (always connected) socketpairs when using the
@@ -762,17 +875,8 @@ class SandboxPolicyCommon : public SandboxPolicyBase {
         if (!mBroker && !mAllowUnsafeSocketPair) {
           return Nothing();
         }
-        // See bug 1066750.
         if (!aHasArgs) {
-          // If this is a socketcall(2) platform, but the kernel also
-          // supports separate syscalls (>= 4.2.0), we can unpack the
-          // arguments and filter them.
-          if (HasSeparateSocketCalls()) {
-            return Some(Trap(SocketpairUnpackTrap, nullptr));
-          }
-          // Otherwise, we can't filter the args if the platform passes
-          // them by pointer.
-          return Some(Allow());
+          return Some(UnpackSocketcallOrAllow());
         }
         Arg<int> domain(0), type(1);
         return Some(
@@ -792,7 +896,7 @@ class SandboxPolicyCommon : public SandboxPolicyBase {
         // Best-effort argument filtering as for socketpair(2), above.
         if (!aHasArgs) {
           if (HasSeparateSocketCalls()) {
-            return Some(Trap(GetSockOptUnpackTrap, nullptr));
+            return Some(Trap(SocketcallUnpackTrap, nullptr));
           }
           return Some(Allow());
         }
@@ -925,12 +1029,17 @@ class SandboxPolicyCommon : public SandboxPolicyBase {
         // filter those; pids do need to be restricted to the current
         // process in order to not leak information.
         Arg<clockid_t> clk_id(0);
+#ifdef MOZ_GECKO_PROFILER
         clockid_t this_process =
             MAKE_PROCESS_CPUCLOCK(getpid(), CPUCLOCK_SCHED);
+#endif
         return If(clk_id == CLOCK_MONOTONIC, Allow())
 #ifdef CLOCK_MONOTONIC_COARSE
             // Used by SandboxReporter, among other things.
             .ElseIf(clk_id == CLOCK_MONOTONIC_COARSE, Allow())
+#endif
+#ifdef CLOCK_MONOTONIC_RAW
+            .ElseIf(clk_id == CLOCK_MONOTONIC_RAW, Allow())
 #endif
             .ElseIf(clk_id == CLOCK_PROCESS_CPUTIME_ID, Allow())
             .ElseIf(clk_id == CLOCK_REALTIME, Allow())
@@ -993,8 +1102,14 @@ class SandboxPolicyCommon : public SandboxPolicyBase {
             .Case(F_GETFL, Allow())
             .Case(F_SETFL, If((flags & ~allowed_flags) == 0, Allow())
                                .Else(InvalidSyscall()))
+#if defined(MOZ_PROFILE_GENERATE)
+            .Case(F_SETLKW, Allow())
+#endif
             // Not much different from other forms of dup(), and commonly used.
             .Case(F_DUPFD_CLOEXEC, Allow())
+            // Used by Mesa, generally useful, and harmless: tests if
+            // two file descriptors refer to the same file description.
+            .Case(F_DUPFD_QUERY, Allow())
             .Default(SandboxPolicyBase::EvaluateSyscall(sysno));
       }
 
@@ -1052,6 +1167,10 @@ class SandboxPolicyCommon : public SandboxPolicyBase {
         // allowed values here also add them to the GMP sandbox rules.
         return If(advice == MADV_DONTNEED, Allow())
             .ElseIf(advice == MADV_FREE, Allow())
+            // Used by glibc (and maybe someday mozjemalloc).
+            .ElseIf(advice == MADV_GUARD_INSTALL, Allow())
+            .ElseIf(advice == MADV_GUARD_REMOVE, Allow())
+            // Formerly used by mozjemalloc; unclear if current use:
             .ElseIf(advice == MADV_HUGEPAGE, Allow())
             .ElseIf(advice == MADV_NOHUGEPAGE, Allow())
 #ifdef MOZ_ASAN
@@ -1120,6 +1239,22 @@ class SandboxPolicyCommon : public SandboxPolicyBase {
             .Else(PrctlPolicy());
       }
 
+#if defined(GP_PLAT_amd64_linux) && defined(GP_ARCH_amd64) && \
+    defined(MOZ_USING_WASM_SANDBOXING)
+        // arch_prctl
+      case __NR_arch_prctl: {
+        // Bug 1923701 - Needed for by RLBox-wasm2c: Buggy libraries are
+        // sandboxed with RLBox and wasm2c (Wasm). wasm2c offers an optimization
+        // for performance that uses the otherwise-unused GS register on x86.
+        // The GS register is only settable using the arch_prctl platforms on
+        // older x86 CPUs that don't have the wrgsbase instruction. This
+        // optimization is currently only supported on linux+clang+x86_64.
+        Arg<int> op(0);
+        return If(op == ARCH_SET_GS, Allow())
+            .Else(SandboxPolicyBase::EvaluateSyscall(sysno));
+      }
+#endif
+
         // NSPR can call this when creating a thread, but it will accept a
         // polite "no".
       case __NR_getpriority:
@@ -1297,9 +1432,6 @@ class ContentSandboxPolicy : public SandboxPolicyCommon {
   Maybe<ResultExpr> EvaluateSocketCall(int aCall,
                                        bool aHasArgs) const override {
     switch (aCall) {
-      case SYS_SENDMMSG:  // libresolv via libasyncns; see bug 1355274
-        return Some(Allow());
-
 #ifdef ANDROID
       case SYS_SOCKET:
         return Some(Error(EACCES));
@@ -1442,9 +1574,10 @@ class ContentSandboxPolicy : public SandboxPolicyCommon {
         return If(request == FIOCLEX, Allow())
             // Rust's stdlib also uses FIONBIO instead of equivalent fcntls.
             .ElseIf(request == FIONBIO, Allow())
-            // Allow anything that isn't a tty ioctl, for now; bug 1302711
-            // will cover changing this to a default-deny policy.
-            .ElseIf(shifted_type != kTtyIoctls, Allow())
+            // Allow anything that isn't a tty ioctl, if level < 6
+            .ElseIf(
+                BelowLevel(6) ? shifted_type != kTtyIoctls : BoolConst(false),
+                Allow())
             .Else(SandboxPolicyCommon::EvaluateSyscall(sysno));
       }
 
@@ -1600,6 +1733,10 @@ class ContentSandboxPolicy : public SandboxPolicyCommon {
         // usually do something reasonable on error.
       case __NR_clone:
         return ClonePolicy(Error(EPERM));
+#  ifdef __NR_fork
+      case __NR_fork:
+        return Error(ENOSYS);
+#  endif
 
 #  ifdef __NR_fadvise64
       case __NR_fadvise64:
@@ -1634,11 +1771,6 @@ class ContentSandboxPolicy : public SandboxPolicyCommon {
 #endif
         return Allow();
 
-#ifdef MOZ_JPROF
-      case __NR_setitimer:
-        return Allow();
-#endif  // MOZ_JPROF
-
       default:
         return SandboxPolicyCommon::EvaluateSyscall(sysno);
     }
@@ -1910,19 +2042,33 @@ class RDDSandboxPolicy final : public SandboxPolicyCommon {
         static constexpr unsigned long kVideoType =
             static_cast<unsigned long>('V') << _IOC_TYPESHIFT;
 #endif
-        // nvidia uses some ioctls from this range (but not actual
+        // nvidia non-tegra uses some ioctls from this range (but not actual
         // fbdev ioctls; nvidia uses values >= 200 for the NR field
         // (low 8 bits))
         static constexpr unsigned long kFbDevType =
             static_cast<unsigned long>('F') << _IOC_TYPESHIFT;
 
+#if defined(__aarch64__)
+        // NVIDIA decoder, from Linux4Tegra
+        // http://lists.mplayerhq.hu/pipermail/ffmpeg-devel/2024-May/328552.html
+        static constexpr unsigned long kNvidiaNvmapType =
+            static_cast<unsigned long>('N') << _IOC_TYPESHIFT;
+        static constexpr unsigned long kNvidiaNvhostType =
+            static_cast<unsigned long>('H') << _IOC_TYPESHIFT;
+#endif  // defined(__aarch64__)
+
         // Allow DRI and DMA-Buf for VA-API. Also allow V4L2 if enabled
         return If(shifted_type == kDrmType, Allow())
             .ElseIf(shifted_type == kDmaBufType, Allow())
 #ifdef MOZ_ENABLE_V4L2
             .ElseIf(shifted_type == kVideoType, Allow())
 #endif
-            // Hack for nvidia, which isn't supported yet:
+        // NVIDIA decoder from Linux4Tegra, this is specific to Tegra ARM64 SoC
+#if defined(__aarch64__)
+            .ElseIf(shifted_type == kNvidiaNvmapType, Allow())
+            .ElseIf(shifted_type == kNvidiaNvhostType, Allow())
+#endif  // defined(__aarch64__)
+        // Hack for nvidia non-tegra devices, which isn't supported yet:
             .ElseIf(shifted_type == kFbDevType, Error(ENOTTY))
             .Else(SandboxPolicyCommon::EvaluateSyscall(sysno));
       }
@@ -1978,6 +2124,10 @@ class RDDSandboxPolicy final : public SandboxPolicyCommon {
         // nvidia drivers may attempt to spawn nvidia-modprobe
       case __NR_clone:
         return ClonePolicy(Error(EPERM));
+#ifdef __NR_fork
+      case __NR_fork:
+        return Error(ENOSYS);
+#endif
 
         // Pass through the common policy.
       default:
@@ -1995,8 +2145,15 @@ UniquePtr<sandbox::bpf_dsl::Policy> GetDecoderSandboxPolicy(
 // Basically a clone of RDDSandboxPolicy until we know exactly what
 // the SocketProcess sandbox looks like.
 class SocketProcessSandboxPolicy final : public SandboxPolicyCommon {
+ private:
+  SocketProcessSandboxParams mParams;
+
+  bool BelowLevel(int aLevel) const { return mParams.mLevel < aLevel; }
+
  public:
-  explicit SocketProcessSandboxPolicy(SandboxBrokerClient* aBroker) {
+  explicit SocketProcessSandboxPolicy(SandboxBrokerClient* aBroker,
+                                      SocketProcessSandboxParams&& aParams)
+      : mParams(std::move(aParams)) {
     mBroker = aBroker;
     mMayCreateShmem = true;
   }
@@ -2016,6 +2173,17 @@ class SocketProcessSandboxPolicy final : public SandboxPolicyCommon {
     }
   }
 
+  BoolExpr MsgFlagsAllowed(const Arg<int>& aFlags) const override {
+    // Necko might use advanced networking features, and the sandbox
+    // is relatively permissive compared to content, so this is a
+    // default-allow policy.
+    //
+    // However, `MSG_OOB` has historically been buggy, and the way it
+    // maps to TCP is notoriously broken (see RFC 6093), so it should
+    // be safe to block.
+    return (aFlags & MSG_OOB) == 0;
+  }
+
   Maybe<ResultExpr> EvaluateSocketCall(int aCall,
                                        bool aHasArgs) const override {
     switch (aCall) {
@@ -2024,9 +2192,17 @@ class SocketProcessSandboxPolicy final : public SandboxPolicyCommon {
       case SYS_BIND:
         return Some(Allow());
 
-        // FIXME(bug 1641401) do we really need this?
+      // sendmsg and recvmmsg needed for HTTP3/QUIC UDP IO. Note sendmsg is
+      // allowed in SandboxPolicyCommon.
+      case SYS_RECVMMSG:
+      // Required for the DNS Resolver thread.
       case SYS_SENDMMSG:
-        return Some(Allow());
+        if (aHasArgs) {
+          Arg<int> flags(3);
+          return Some(
+              If(MsgFlagsAllowed(flags), Allow()).Else(InvalidSyscall()));
+        }
+        return Some(UnpackSocketcallOrAllow());
 
       case SYS_GETSOCKOPT:
       case SYS_SETSOCKOPT:
@@ -2052,6 +2228,9 @@ class SocketProcessSandboxPolicy final : public SandboxPolicyCommon {
                 PR_SET_DUMPABLE,  // Crash reporting
                 PR_SET_PTRACER),  // Debug-mode crash handling
                Allow())
+#if defined(MOZ_PROFILE_GENERATE)
+        .CASES((PR_GET_PDEATHSIG), Allow())
+#endif  // defined(MOZ_PROFILE_GENERATE)
         .Default(InvalidSyscall());
   }
 
@@ -2065,15 +2244,21 @@ class SocketProcessSandboxPolicy final : public SandboxPolicyCommon {
         auto shifted_type = request & kIoctlTypeMask;
 
         // Rust's stdlib seems to use FIOCLEX instead of equivalent fcntls.
-        return If(request == FIOCLEX, Allow())
+        return Switch(request)
+            .Case(FIOCLEX, Allow())
             // Rust's stdlib also uses FIONBIO instead of equivalent fcntls.
-            .ElseIf(request == FIONBIO, Allow())
+            .Case(FIONBIO, Allow())
             // This is used by PR_Available in nsSocketInputStream::Available.
-            .ElseIf(request == FIONREAD, Allow())
-            // Allow anything that isn't a tty ioctl, for now; bug 1302711
-            // will cover changing this to a default-deny policy.
-            .ElseIf(shifted_type != kTtyIoctls, Allow())
-            .Else(SandboxPolicyCommon::EvaluateSyscall(sysno));
+            .Case(FIONREAD, Allow())
+            // WebRTC needs interface information (bug 1975576)
+            .CASES((SIOCGIFNAME, SIOCGIFFLAGS, SIOCETHTOOL, SIOCGIWRATE),
+                   Allow())
+            .Default(
+                // Allow anything that isn't a tty ioctl (if level < 2)
+                If(BelowLevel(2) ? shifted_type != kTtyIoctls
+                                 : BoolConst(false),
+                   Allow())
+                    .Else(SandboxPolicyCommon::EvaluateSyscall(sysno)));
       }
 
       CASES_FOR_fcntl: {
@@ -2124,9 +2309,9 @@ class SocketProcessSandboxPolicy final : public SandboxPolicyCommon {
 };
 
 UniquePtr<sandbox::bpf_dsl::Policy> GetSocketProcessSandboxPolicy(
-    SandboxBrokerClient* aMaybeBroker) {
+    SandboxBrokerClient* aMaybeBroker, SocketProcessSandboxParams&& aParams) {
   return UniquePtr<sandbox::bpf_dsl::Policy>(
-      new SocketProcessSandboxPolicy(aMaybeBroker));
+      new SocketProcessSandboxPolicy(aMaybeBroker, std::move(aParams)));
 }
 
 class UtilitySandboxPolicy : public SandboxPolicyCommon {
@@ -2148,6 +2333,12 @@ class UtilitySandboxPolicy : public SandboxPolicyCommon {
                 PR_GET_PDEATHSIG),  // PGO profiling, cf
                                     // https://reviews.llvm.org/D29954
                Allow())
+        .CASES((PR_CAPBSET_READ),  // libcap.so.2 loaded by libpulse.so.0
+                                   // queries for capabilities
+               Error(EINVAL))
+#if defined(MOZ_PROFILE_GENERATE)
+        .CASES((PR_GET_PDEATHSIG), Allow())
+#endif  // defined(MOZ_PROFILE_GENERATE)
         .Default(InvalidSyscall());
   }
 
